# SPDX-FileCopyrightText: 2023-present rmrighes <rmrighes@users.noreply.github.com>
#
# SPDX-License-Identifier: MIT
"""File __init__.py responsible for enabling the import of yajaw package and
global configuration settings."""

from enum import Enum

from yajaw.configuration import YajawConfig

__all__ = ["jira", "configuration", "ApiType"]

ApiType = Enum("API", ["CLASSIC", "AGILE", "INTERNAL"])
Option = Enum("Confirmation", ["YES", "NO"])

YajawConfig.load_settings()



"""TBD"""
import asyncio
import logging
import tomllib
from pathlib import Path
from typing import ClassVar


class YajawConfig:
    """TBD"""

    JIRA_PAT: str
    JIRA_BASE_URL: str
    SERVER_API_V2: str
    SERVER_API: str
    AGILE_API_V1: str
    AGILE_API: str
    GREENHOPPER_API: str
    TRIES: int
    DELAY: float
    BACKOFF: float
    LOGGER: logging.Logger
    SEMAPHORE: asyncio.BoundedSemaphore
    TIMEOUT: int
    DEFAULT_PAGINATION: dict

    _MIN_SEMAPHORE_LIMIT: int = 5

    __conf: ClassVar = {
        "jira": {
            "token": "AAA...",
            "base_url": "https://my-dummy-jira-url.com",
            "server_api_v2": "rest/api/2",
            "agile_api_v1": "rest/agile/1.0",
            "greenhopper_api": "rest/greenhopper/1.0",
        },
        "log": {"msg_format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"},
        "retries": {"tries": 10, "delay": 0.0, "backoff": 2},
        "requests": {"timeout": 60},
        "concurrency": {"semaphore_limit": 50},
        "pagination": {"page_results": 40},
    }

    __sections: ClassVar = ["jira", "log", "retries", "requests", "concurrency"]

    @staticmethod
    def configuration(section, setting):
        """TBD"""
        return YajawConfig.__conf[section][setting]

    @staticmethod
    def update_configuration(section, setting, value):
        """TBD"""
        if section in YajawConfig.__sections:
            YajawConfig.__conf[section][setting] = value
        else:
            raise NameError

    @staticmethod
    def load_settings():
        """Load configuration settings from file"""
        fname = "yajaw.toml"
        try:
            with open(Path.home() / ".yajaw" / fname, "rb") as toml:
                YajawConfig.__conf = tomllib.load(toml)
        except FileNotFoundError:
            ...
        YajawConfig.__conf["log"]["logger"] = YajawConfig.define_logger()
        limit = YajawConfig.__conf["concurrency"]["semaphore_limit"]
        # Ensures da minimum value of 5 for the BoundedSemaphore
        semaphore_limit = (
            limit if limit > YajawConfig._MIN_SEMAPHORE_LIMIT else YajawConfig._MIN_SEMAPHORE_LIMIT
        )
        YajawConfig.__conf["concurrency"]["semaphore"] = asyncio.BoundedSemaphore(semaphore_limit)
        YajawConfig.__conf["pagination"]["default"] = {
            "startAt": 0,
            "maxResults": YajawConfig.__conf["pagination"]["page_results"],
        }
        YajawConfig.set_class_variables()

    @staticmethod
    def define_logger():
        """Configure the global settings for logging."""
        logging.getLogger("httpx").setLevel(logging.WARNING)
        logging.basicConfig(level=logging.INFO, format=YajawConfig.__conf["log"]["msg_format"])
        return logging.getLogger(__package__)

    @staticmethod
    def set_class_variables():
        """TBD"""
        YajawConfig.JIRA_PAT = YajawConfig.__conf["jira"]["token"]
        YajawConfig.JIRA_BASE_URL = YajawConfig.__conf["jira"]["base_url"]
        YajawConfig.SERVER_API_V2 = YajawConfig.__conf["jira"]["server_api_v2"]
        YajawConfig.SERVER_API = YajawConfig.SERVER_API_V2
        YajawConfig.AGILE_API_V1 = YajawConfig.__conf["jira"]["agile_api_v1"]
        YajawConfig.AGILE_API = YajawConfig.AGILE_API_V1
        YajawConfig.GREENHOPPER_API = YajawConfig.__conf["jira"]["greenhopper_api"]
        YajawConfig.TRIES = YajawConfig.__conf["retries"]["tries"]
        YajawConfig.DELAY = YajawConfig.__conf["retries"]["delay"]
        YajawConfig.BACKOFF = YajawConfig.__conf["retries"]["backoff"]
        YajawConfig.LOGGER = YajawConfig.__conf["log"]["logger"]
        YajawConfig.SEMAPHORE = YajawConfig.__conf["concurrency"]["semaphore"]
        YajawConfig.TIMEOUT = YajawConfig.__conf["requests"]["timeout"]
        YajawConfig.DEFAULT_PAGINATION = YajawConfig.__conf["pagination"]["default"]



"""Module responsible for lower level HTTP requests."""
import asyncio
import math
import secrets
from http import HTTPStatus

import httpx

from yajaw import Option, YajawConfig
from yajaw.core import exceptions


class PersonalAccessTokenAuth(httpx.Auth):
    """Class used for Personal Access Token auth in httpx."""

    def __init__(self, pat):
        self.token = pat

    def auth_flow(self, request):
        request.headers["Authorization"] = f"Bearer {self.token}"
        yield request


def _generate_headers() -> dict:
    """Function responsible for generating the headers info for HTTP requests."""
    return {"Accept": "application/json", "Content-Type": "application/json"}


def _generate_params(new_params: dict, existing_params: dict | None = None) -> dict:
    """Function responsible for generating the parameters info for HTTP requests."""
    if existing_params is None:
        existing_params = {}
    return existing_params | new_params


def _generate_payload(new_content: dict, existing_content: dict | None = None) -> dict:
    """Function responsible for generating the payload info for HTTP requests."""
    if existing_content is None:
        existing_content = {}
    return existing_content | new_content


def _generate_url(resource: str, api: str) -> str:
    """Function responsible for generating the url info for HTTP requests."""
    return f"{YajawConfig.JIRA_BASE_URL}/{api}/{resource}"


def _generate_auth() -> PersonalAccessTokenAuth:
    """Function responsible for generating the authentication info for HTTP requests."""
    return PersonalAccessTokenAuth(YajawConfig.JIRA_PAT)


def _generate_client() -> httpx.AsyncClient:
    """Function responsible for generating the client used in the context for HTTP requests."""
    return httpx.AsyncClient(
        auth=_generate_auth(),
        headers=_generate_headers(),
        timeout=YajawConfig.TIMEOUT,
        follow_redirects=True,
    )


class JiraInfo:
    """TBD"""

    def __init__(
        self,
        info: dict,
    ):
        self.__method = info["method"]
        self.__resource = info["resource"]
        self.__api = info["api"]
        self.__url = _generate_url(resource=info["resource"], api=info["api"])
        self.__params = info["params"]
        self.__payload = info["payload"]

    @property
    def method(self):
        """TBD"""
        return self.__method

    @method.setter
    def method(self, mhd: str):
        """TBD"""
        self.__method = mhd

    @property
    def resource(self):
        """TBD"""
        return self.__resource

    @resource.setter
    def resource(self, rce: str):
        """TBD"""
        self.__resource = rce

    @property
    def api(self):
        """TBD"""
        return self.__api

    @api.setter
    def api(self, api: str):
        """TBD"""
        self.__api = api

    @property
    def url(self):
        """TBD"""
        return self.__url

    @url.setter
    def url(self, url: str):
        """TBD"""
        self.__url = url

    @property
    def params(self):
        """TBD"""
        return self.__params

    @params.setter
    def params(self, pms: dict):
        """TBD"""
        self.__params = pms

    @property
    def payload(self):
        """TBD"""
        return self.__payload

    @payload.setter
    def payload(self, pad: dict):
        """TBD"""
        self.__payload = pad


def _retry_response_error_detected(result: httpx.Response) -> bool:
    """Check if a retry should proceed based on the HTTP response."""
    retry = True
    if not isinstance(result, httpx.Response):
        _log_and_raise_request_issue("InvalidResponseError")
    if result.status_code == HTTPStatus.UNAUTHORIZED:
        _log_and_raise_request_issue("ResourceUnauthorizedError")
    if result.status_code == HTTPStatus.FORBIDDEN:
        _log_and_raise_request_issue("ResourceForbiddenError")
    if result.status_code == HTTPStatus.METHOD_NOT_ALLOWED:
        _log_and_raise_request_issue("ResourceMethodNotAllowedError")
    if result.status_code == HTTPStatus.NOT_FOUND:
        _log_and_raise_request_issue("ResourceNotFoundError")
    if httpx.codes.is_success(result.status_code):
        retry = False
    return retry


def _log_and_raise_request_issue(error_type: str, warning: Option = Option.NO) -> None:
    """Log an error or warning message and raise the specified exception."""
    log_level = YajawConfig.LOGGER.warning if warning is Option.YES else YajawConfig.LOGGER.error
    log_message = f"log message -- {error_type}"
    log_level(log_message)
    raise getattr(exceptions, error_type)


async def _retry_request(jira: JiraInfo, client: httpx.AsyncClient):
    """Retry the given function on certain conditions."""
    delay = secrets.SystemRandom().uniform(0, YajawConfig.DELAY)
    for attempt in range(1, YajawConfig.TRIES + 1):
        await asyncio.sleep(delay)
        result = await _send_request(jira=jira, client=client)
        _log_attempt_info(result, attempt, delay, error=Option.NO)
        if not _retry_response_error_detected(result):
            return result
        delay *= YajawConfig.BACKOFF
    _log_attempt_info(result, attempt, delay, error=Option.YES)
    raise exceptions.InvalidResponseError


def _log_attempt_info(result, attempt, delay, error=Option.NO):
    """Log information for a retry attempt."""
    log_level = YajawConfig.LOGGER.error if error == Option.YES else YajawConfig.LOGGER.info

    if isinstance(result, httpx.Response):
        log_message = (
            f"{result.status_code} - attempt {attempt:02d} - "
            f"delay {delay:05.2f}s -- {result.request.method} "
            f"{result.request.url}"
        )
    else:
        log_message = f"Non-HTTPX response - attempt {attempt:02d} - delay {delay:04.2f}s"

    log_level(log_message)


async def _send_request(jira: JiraInfo, client: httpx.AsyncClient) -> httpx.Response:
    """Function responsible for making a low-level HTTP request."""
    method, url, params, payload = jira.method, jira.url, jira.params, jira.payload
    async with YajawConfig.SEMAPHORE:
        return await client.request(method=method, url=url, params=params, json=payload)


async def send_single_request(
    jira: JiraInfo, client: httpx.AsyncClient | None = None
) -> httpx.Response:
    """Function wraps retry_request and returns a list of responses."""
    try:
        if client is None:
            client = _generate_client()
        task = asyncio.create_task(_retry_request(jira=jira, client=client))
        response = await task
    except exceptions.ResourceNotFoundError as exc:
        YajawConfig.LOGGER.warning("pending log message -- send_single_request")
        raise exceptions.ResourceNotFoundError from exc
    except exceptions.YajawError as e:
        YajawConfig.LOGGER.exception("pending log message -- send_single_request")
        raise exceptions.YajawError from e
    return response


async def send_paginated_requests(jira: JiraInfo) -> list[httpx.Response]:
    """Function that wraps send_single_request and and call it for all necessary pages."""
    default_page_attr = YajawConfig.DEFAULT_PAGINATION
    jira_list = _create_jira_list_with_page_attr(page_attr_list=[default_page_attr], jira=jira)
    initial_jira = jira_list[0]

    client = _generate_client()
    async with client:
        # First request with default pagination
        response = await send_single_request(jira=initial_jira, client=client)

        responses = []
        responses.append(response)

        # Identify if additional requests are needed
        page_attr = _retrieve_pagination_attributes(response=response)
        if _is_pagination_required(page_attr=page_attr):
            # Generate the updated page_attr_list
            page_attr_list = _create_list_of_page_attr(page_attr=page_attr)

            # Generate the updated jira_list
            jira_list = _create_jira_list_with_page_attr(page_attr_list=page_attr_list, jira=jira)

            # Create concurrent requests for the additional pages
            # iterate over jira_list
            # Consolidate responses list
            # for response in paginated_responses: responses.extend(response)

            tasks = [
                asyncio.create_task(send_single_request(jira=jira, client=client))
                for jira in jira_list
            ]
            group_of_response = await asyncio.gather(*tasks)

            responses.extend(list(group_of_response))

    return responses


def _create_list_of_page_attr(page_attr: dict) -> list[dict]:
    """Function that generates a list of attributes needed to retrieve all pages."""
    if "total" not in page_attr:
        return [page_attr]

    page_attr["start_at"] = page_attr["max_results"]

    page_attr_list = [
        page_attr["start_at"] + i * page_attr["max_results"]
        for i in range(int(math.ceil(page_attr["total"] / page_attr["max_results"])) - 1)
    ]
    return [{"startAt": page, "maxResults": page_attr["max_results"]} for page in page_attr_list]


def _create_jira_list_with_page_attr(page_attr_list: list[dict], jira: JiraInfo) -> list[JiraInfo]:
    """Function that gets a list of page attributes and createa
    a list with respective JiraInfo objects."""
    jira_list = []
    for page_attr in page_attr_list:
        unique_info = {
            "method": jira.method,
            "resource": jira.resource,
            "api": jira.api,
            "params": jira.params,
            "payload": jira.payload,
        }
        if unique_info["method"] == "GET":
            unique_info["params"] = _generate_params(
                new_params=page_attr, existing_params=unique_info["params"]
            )
        if unique_info["method"] == "POST":
            unique_info["payload"] = _generate_payload(
                new_content=page_attr, existing_content=unique_info["payload"]
            )
        unique_jira = JiraInfo(info=unique_info)
        jira_list.append(unique_jira)
    return jira_list


def _retrieve_pagination_attributes(response: httpx.Response) -> dict:
    """TBD"""
    page_attr = {}
    response_json = response.json()
    page_attr["start_at"] = response_json["startAt"] if "startAt" in response_json else None
    page_attr["max_results"] = (
        response_json["maxResults"] if "maxResults" in response_json else None
    )
    page_attr["total"] = response_json["total"] if "total" in response_json else None
    return page_attr


def _is_pagination_required(page_attr: dict) -> bool:
    """Function that checks if pagination is required."""
    start_at = page_attr["start_at"]
    max_results = page_attr["max_results"]
    total = page_attr["total"]
    return start_at + max_results < total



"""Module responsonsible for testing yajaw.jira module."""
from unittest.mock import patch

import httpx
import pytest

from yajaw import jira
from yajaw.core import exceptions as e

#########
# MOCK!!!
#########

# Reference for httpx.Response:
# https://github.com/encode/httpx/blob/master/httpx/_models.py

# status_code: int,
# headers: typing.Optional[HeaderTypes] = None,
# content: typing.Optional[ResponseContent] = None,
# text: typing.Optional[str] = None,
# html: typing.Optional[str] = None,
# json: typing.Any = None,
# stream: typing.Union[SyncByteStream, AsyncByteStream, None] = None,
# request: typing.Optional[Request] = None,
# extensions: typing.Optional[ResponseExtensions] = None,
# history: typing.Optional[typing.List["Response"]] = None,
# default_encoding: typing.Union[str, typing.Callable[[bytes], str]] = "utf-8",


def fetch_single_mock_valid_issue() -> httpx.Response:
    """Auxiliry function to generate a valid issue as httpx.Response to be
    used in different tests.
    """
    return httpx.Response(
        status_code=200,
        headers=None,
        request=httpx.Request("GET", "https://example.org"),
        json={"key": "VALID"},
    )


def fetch_single_mock_server_not_found() -> httpx.Response:
    """Auxiliary function to generate a server not found httpx.Response
    to be used in different tests.
    """
    return httpx.Response(status_code=503, headers=None, request=None, json={})


def fetch_single_mock_not_found_issue() -> httpx.Response:
    """Auxiliary function to generate a resource not found httpx.Response
    to be used in different tests.
    """
    return httpx.Response(status_code=404, headers=None, request=None, json={})


@patch("httpx.AsyncClient.request")
def test_fetch_single_valid_issue(mock_rest_request):
    """Test fetch_project() with a single valid issue."""
    mock_response = fetch_single_mock_valid_issue()
    mock_rest_request.return_value = mock_response
    response = jira.fetch_project("VALID")
    assert isinstance(response, dict)
    assert response["key"] == "VALID"


@patch("httpx.AsyncClient.request")
@patch("asyncio.sleep")
def test_fetch_single_issue_with_server_not_found(mock_rest_request, mock_sleep):
    """Test fetch_project() with an invalid response."""
    mock_value = fetch_single_mock_server_not_found()
    mock_rest_request.return_value = mock_value
    mock_sleep.return_value = None
    with pytest.raises(e.YajawError):
        jira.fetch_project("INVALID-KEY")



# SPDX-FileCopyrightText: 2023-present rmrighes <rmrighes@users.noreply.github.com>
#
# SPDX-License-Identifier: MIT
"""File __about__.py responsible for yajaw package meta information."""

__version__ = "0.0.2.dev1"



"""TBD"""
import asyncio
from contextlib import contextmanager
from functools import wraps
from time import perf_counter

from yajaw import YajawConfig


def duration(func):
    """TBD"""

    @contextmanager
    def wrapping_logic():
        start_ts = perf_counter()
        yield
        dur = perf_counter() - start_ts
        YajawConfig.LOGGER.info(f"{func.__name__} took {dur:.2f} seconds")

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not asyncio.iscoroutinefunction(func):
            with wrapping_logic():
                return func(*args, **kwargs)
        else:

            async def tmp():
                with wrapping_logic():
                    return await func(*args, **kwargs)

            return tmp()

    return wrapper



"""TBD"""
import asyncio

from yajaw import ApiType, jira
from yajaw.utils.decorators import duration


async def async_main():
    """TBD"""

    await use_async_fetch_all_projects()
    print()
    await use_async_fetch_projects()
    print()
    await use_async_fetch_projects_from_list()
    print()
    await use_async_fetch_issue()
    print()
    await use_async_fetch_issue_from_agile()
    print()
    await use_async_search_issues()
    print()
    # await use_async_fetch_sprints()
    # print()
    # await use_async_fetch_issues_from_sprints()


@duration
async def use_async_fetch_all_projects():
    """TBD"""
    print("ASYNC RESOURCE: PROJECT")
    projects = await jira.async_fetch_all_projects()
    if projects:
        for project in projects:
            print(f"Project key: {project['key']}")
        print(f"There is a total of {len(projects)} projects.")
    else:
        print("empty list!")


@duration
async def use_async_fetch_projects():
    """TBD"""
    print("ASYNC RESOURCE: PROJECT/{KEY}")
    project = await jira.async_fetch_project(project_key="EMI")
    if project:
        print(f"Project key: {project['key']}")
    else:
        print("empty dictionary!")


@duration
async def use_async_fetch_projects_from_list():
    """TBD"""
    print("ASYNC RESOURCE: PROJECT FROM LIST")
    project_keys = ["ETOE", "ESTT", "EMI"]
    projects = await jira.async_fetch_projects_from_list(project_keys=project_keys)
    if projects:
        for project in projects:
            print(f"Project key: {project['key']}")
        print(f"There is a total of {len(projects)} projects.")
    else:
        print("empty list!")


@duration
async def use_async_fetch_issue():
    """TBD"""
    print("ASYNC RESOURCE: ISSUE/{KEY}")
    issue = await jira.async_fetch_issue(issue_key="ETOE-8")
    if issue:
        print(f"Issue key: {issue['key']}")
    else:
        print("empty dictionary!")


@duration
async def use_async_fetch_issue_from_agile():
    """TBD"""
    print("ASYNC AGILE RESOURCE: ISSUE/{KEY}")
    issue = await jira.async_fetch_issue(
        issue_key="MLOPS-8724", agile=ApiType.AGILE, expand="changelog"
    )
    if issue:
        # story_points = "customfield_10012"
        print(f"Issue key: {issue['key']}")
        print(f"Estimate: {issue['fields']['customfield_10012']}")
        print(*(sprint for sprint in issue["fields"]["customfield_10631"]), sep="\n")
    else:
        print("empty dictionary!")


@duration
async def use_async_search_issues():
    """TBD"""
    print("ASYNC RESOURCE: SEARCH")
    jql = "project in (ETOE, ESTT, EMI, MLOPS)"
    issues = await jira.async_search_issues(jql=jql)
    if issues:
        for issue in issues:
            print(f"{issue['key']} - {issue['fields']['summary']}")
        print(f"There is a total of {len(issues)} projects.")
    else:
        print("empty list!")


@duration
async def use_async_fetch_sprints():
    """TBD"""
    print("ASYNC RESOURCE: sprint")
    # Continue implementation


@duration
async def use_async_fetch_issues_from_sprints():
    """TBD"""
    # Continue implementation


def sync_main():
    """TBD"""
    use_fetch_all_projects()
    print()
    use_fetch_projects()
    print()
    use_fetch_projects_from_list()
    print()
    use_fetch_issue()
    print()
    use_fetch_issue_from_agile()
    print()
    use_search_issues()
    print()
    # use_fetch_sprints()
    # print()
    # use_fetch_issues_from_sprints()


@duration
def use_fetch_all_projects():
    """TBD"""
    print("RESOURCE: PROJECT")
    projects = jira.fetch_all_projects()
    if projects:
        for project in projects:
            print(f"Project key: {project['key']}")
        print(f"There is a total of {len(projects)} projects.")
    else:
        print("empty list!")


@duration
def use_fetch_projects():
    """TBD"""
    print("RESOURCE: PROJECT/{KEY}")
    project = jira.fetch_project(project_key="EMI")
    if project:
        print(f"Project key: {project['key']}")
    else:
        print("empty dictionary!")


@duration
def use_fetch_projects_from_list():
    """TBD"""
    print("RESOURCE: PROJECT FROM LIST")
    project_keys = ["ETOE", "ESTT", "EMI"]
    projects = jira.fetch_projects_from_list(project_keys=project_keys)
    if projects:
        for project in projects:
            print(f"Project key: {project['key']}")
        print(f"There is a total of {len(projects)} projects.")
    else:
        print("empty list!")


@duration
def use_fetch_issue():
    """TBD"""
    print("RESOURCE: ISSUE/{KEY}")
    issue = jira.fetch_issue(issue_key="ETOE-8")
    if issue:
        print(f"Issue key: {issue['key']}")
    else:
        print("empty dictionary!")


@duration
def use_fetch_issue_from_agile():
    """TBD"""
    print("ASYNC AGILE RESOURCE: ISSUE/{KEY}")
    issue = jira.fetch_issue(issue_key="MLOPS-8730", expand="changelog", agile=ApiType.AGILE)
    if issue:
        # story_points = "customfield_10012"
        print(f"Issue key: {issue['key']}")
        print(f"Estimate: {issue['fields']['customfield_10012']}")
        print(*(sprint for sprint in issue["fields"]["customfield_10631"]), sep="\n")
    else:
        print("empty dictionary!")


@duration
def use_search_issues():
    """TBD"""
    print("RESOURCE: SEARCH")
    jql = "project in (ETOE, ESTT, EMI, MLOPS)"
    issues = jira.search_issues(jql=jql)
    if issues:
        for issue in issues:
            print(f"{issue['key']} - {issue['fields']['summary']}")
        print(f"There is a total of {len(issues)} projects.")
    else:
        print("empty list!")


@duration
def use_fetch_sprints():
    """TBD"""
    print("ASYNC RESOURCE: sprint")
    # Continue implementation


@duration
def use_fetch_issues_from_sprints():
    """TBD"""
    # Continue implementation


if __name__ == "__main__":
    CONCURRENT = False

    if not CONCURRENT:
        sync_main()
    else:
        raise SystemExit(asyncio.run(async_main()))



""" Module responsible for the definition of custom exceptions."""


class YajawError(Exception):
    "Base class for yajaw errors"


class InvalidResponseError(YajawError):
    "Response is not valid!"


class HttpClientError(YajawError):
    "Returned HTTP Status Code is 4xx"


class HttpServerError(YajawError):
    "Returned HTTP Status Code is 5xx"


class ResourceUnauthorizedError(HttpClientError):
    "Resource unauthorized"


class ResourceForbiddenError(HttpClientError):
    "Resource forbidden"


class ResourceNotFoundError(HttpClientError):
    "Resource not found"


class ResourceMethodNotAllowedError(HttpClientError):
    "Resource does not support method"



"""TBD"""
import asyncio

from yajaw import ApiType, YajawConfig
from yajaw.core import exceptions as e
from yajaw.core import rest


async def async_fetch_all_projects(expand: str | None = None) -> list[dict]:
    """TBD"""
    expand_dict = {} if expand is None else {"expand": expand}

    info = {
        "method": "GET",
        "resource": "project",
        "api": YajawConfig.SERVER_API,
        "params": expand_dict,
        "payload": None,
    }

    jira = rest.JiraInfo(info=info)

    try:
        response = await rest.send_single_request(jira=jira)
        return response.json()
    except e.ResourceNotFoundError:
        return []


def fetch_all_projects(expand: str | None = None) -> list[dict]:
    """Wrapper sync function for GET /project"""
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    coroutine = async_fetch_all_projects(expand=expand)
    return loop.run_until_complete(coroutine)


async def async_fetch_project(project_key: str, expand: str | None = None) -> dict:
    """TBD"""
    expand_dict = {} if expand is None else {"expand": expand}

    info = {
        "method": "GET",
        "resource": f"project/{project_key}",
        "api": YajawConfig.SERVER_API,
        "params": expand_dict,
        "payload": None,
    }

    jira = rest.JiraInfo(info=info)

    try:
        response = await rest.send_single_request(jira=jira)
        return response.json()
    except e.ResourceNotFoundError:
        return {}


def fetch_project(project_key: str, expand: str | None = None) -> dict:
    """Wrapper sync function for GET /project/{key}"""
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    coroutine = async_fetch_project(project_key=project_key, expand=expand)
    return loop.run_until_complete(coroutine)


async def async_fetch_projects_from_list(
    project_keys: list[str], expand: str | None = None
) -> list[dict]:
    """TBD"""
    expand_dict = {} if expand is None else {"expand": expand}

    info_list = [
        {
            "method": "GET",
            "resource": f"project/{project_key}",
            "api": YajawConfig.SERVER_API,
            "params": expand_dict,
            "payload": None,
        }
        for project_key in project_keys
    ]

    jira_list = [rest.JiraInfo(info=info) for info in info_list]

    try:
        tasks = [asyncio.create_task(rest.send_single_request(jira=jira)) for jira in jira_list]
        responses = await asyncio.gather(*tasks)
        return [response.json() for response in responses]
    except e.ResourceNotFoundError:
        return []


def fetch_projects_from_list(project_keys: list[str], expand: str | None = None) -> list[dict]:
    """Wrapper sync function for multiple calls on GET /project/{key}"""
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    coroutine = async_fetch_projects_from_list(project_keys=project_keys, expand=expand)
    return loop.run_until_complete(coroutine)


async def async_fetch_issue(
    issue_key: str, expand: str | None = None, agile: ApiType = ApiType.CLASSIC
) -> dict:
    """TBD"""
    expand_dict = {} if expand is None else {"expand": expand}

    api = YajawConfig.SERVER_API if not agile else YajawConfig.AGILE_API

    info = {
        "method": "GET",
        "resource": f"issue/{issue_key}",
        "api": api,
        "params": expand_dict,
        "payload": None,
    }

    jira = rest.JiraInfo(info=info)

    try:
        response = await rest.send_single_request(jira=jira)
        return response.json()
    except e.ResourceNotFoundError:
        return {}


def fetch_issue(
    issue_key: str, expand: str | None = None, agile: ApiType = ApiType.CLASSIC
) -> dict:
    """Wrapper sync function for GET /issue/{key}"""
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    coroutine = async_fetch_issue(issue_key=issue_key, expand=expand, agile=agile)
    return loop.run_until_complete(coroutine)


async def async_search_issues(jql: str, expand: str | None = None) -> list[dict]:
    """TBD"""
    expand_dict = {} if expand is None else {"expand": expand}
    query = {"jql": jql}

    info = {
        "method": "POST",
        "resource": "search",
        "api": YajawConfig.SERVER_API,
        "params": expand_dict,
        "payload": query,
    }

    jira = rest.JiraInfo(info=info)

    try:
        responses = await rest.send_paginated_requests(jira=jira)
        return [issue for issue_page in responses for issue in issue_page.json()["issues"]]
    except e.ResourceNotFoundError:
        return []


def search_issues(jql: str, expand: str | None = None) -> list[dict]:
    """Wrapper sync function for POST /search"""
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    coroutine = async_search_issues(jql=jql, expand=expand)
    return loop.run_until_complete(coroutine)
